(* Delphi Program
   Generate digital fingerprints for files in a directory tree
   Multithreaded version

    2011-2020 J. Rathlev 24222 Schwentinental
      Web:  www.rathlev-home.de
      Mail: kontakt(a)rathlev-home.de

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1.0 - December 2011
         1.1 - June 2012    : speeded up version with optimized routines
         3.0 - March 2018   : use functions from System.Hash
                              added: SHA256
         3.1 - July 2018    : cosmetic changes
         3.2 - January 2020 : improved batch call (automatic close)

   Calling:
     FileFP.exe [filename] [/md5] [/sha1] [/sha256] [/out] [/force]
     Options: /md5, /sha1, /sha256 : hashes to be calculated
              /out                 : write hashes to file (filename.txt)
              /force               : start calculation immediately
     if both options /out and /force are specified, the program closes automatically
     (e.g. for batch processing)
   *)

unit CalcSigMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes,
  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.Buttons,
  HListBox, Vcl.ExtCtrls, Vcl.ComCtrls, CBFunctions, Vcl.ImgList, Vcl.CheckLst,
  System.ImageList;

const
  ProgName = 'Generate digital fingerprints for files';
  Vers = ' - Vers. 3.2';
  CopRgt = ' 2015-2020 Dr. J. Rathlev, D-24222 Schwentinental';
  EmailAdr = 'kontakt(a)rathlev-home.de';

  defOutfile = 'fingerprints.txt';
  TxtExt = 'txt';
  defBufSize = 1024*1024*32;

type
  TSigType = (stMD5,stSHA1,stSHA256);
  TSigTypes = set of TSigType;

const
  nThreads = 3;
  SigNames : array [TSigType] of string = ('MD5','SHA1','SHA256');

type
  TSignatures = array [TSigType] of AnsiString;

  TDigiSigThread = class(TThread)
  private
    FActiveSig      : TSigType;
    FSignature,
    FFilename       : string;
    FStream         : TFileStream;
    FError          : integer;
    FProgressType   : TProgressType;
    FUserBreak      : boolean;
    FSize,
    FPosition,
    FProgressValue  : int64;
    FOnProgress     : TProgressEvent;
//    procedure SetProgress (ACallBack : TProgressEvent);
    function GetPosition : integer;
    function GetDone : boolean;
  protected
    function OpenFile : integer;
    function ProcessFile : integer; virtual;
    procedure UpdateProgress;
    procedure DoProgress (Progresstype : TProgressType; Value : int64);
    procedure Execute; override;
  public
    constructor Create (const AFilename : string; ASignatureType : TSigType; ASuspend : Boolean);
    procedure CancelThread; virtual;
    property Filename : string read FFilename;
    property ProcessStatus : integer read GetPosition;  // 0 .. 100
    property Done : boolean read GetDone;
    property ErrorCode : integer read FError;
    property Signature : string read FSignature;
    property SignatureType : TSigType read FActiveSig;
    property OnProgress : TProgressEvent read FOnProgress write FOnProgress;// SetProgress;
  end;

  TMD5Thread = class(TDigiSigThread)
  protected
    function ProcessFile : integer; override;
  public
    constructor Create (const AFilename : string; ASuspend : Boolean);
    procedure CancelThread; override;
    end;

  TSHA1Thread = class(TDigiSigThread)
  protected
    function ProcessFile : integer; override;
  public
    constructor Create (const AFilename : string; ASuspend : Boolean);
    procedure CancelThread; override;
    end;

  TSHA256Thread = class(TDigiSigThread)
  protected
    function ProcessFile : integer; override;
  public
    constructor Create (const AFilename : string; ASuspend : Boolean);
    procedure CancelThread; override;
    end;

  TThreads = record
    FilePath     : string;
    FileIndex,
    StatusIndex  : integer;
    Active       : boolean;
    SigThread    : array [TSigType] of TDigiSigThread;
    end;

  TFileSigs = class(TObject)
    Filename : string;
    Size     : int64;
    Signatures : TSignatures;
    constructor Create (AFilename : string; ASize : int64);
    procedure SetSignature( AType : TSigType; ASig : string);
    end;

  TMainForm = class(TForm)
    Label1: TLabel;
    SourceDirEdit: THistoryCombo;
    SourceSelectBtn: TSpeedButton;
    FileBtn: TSpeedButton;
    FileEdit: THistoryCombo;
    Label2: TLabel;
    OutFileEdit: THistoryCombo;
    EndeBtn: TBitBtn;
    InfoBtn: TSpeedButton;
    cbSubDir: TCheckBox;
    OpenDialog: TOpenDialog;
    cbMD5: TCheckBox;
    cbSHA1: TCheckBox;
    ProgressBar: TProgressBar;
    OutBtn: TSpeedButton;
    ImageList: TImageList;
    Label5: TLabel;
    edThreads: TEdit;
    udThreads: TUpDown;
    StatusBar: TStatusBar;
    lvFiles: TListView;
    pnTop: TPanel;
    btnProcess: TBitBtn;
    cbPriority: TComboBox;
    cbOut: TCheckBox;
    pnSettings: TPanel;
    gbSig: TGroupBox;
    cbSepFile: TCheckBox;
    cbSHA256: TCheckBox;
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure EndeBtnClick(Sender: TObject);
    procedure InfoBtnClick(Sender: TObject);
    procedure SourceSelectBtnClick(Sender: TObject);
    procedure FileBtnClick(Sender: TObject);
    procedure btnProcessClick(Sender: TObject);
    procedure FileEditCloseUp(Sender: TObject);
    procedure OutBtnClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    { Private-Deklarationen }
    ProgVersName,
    ProgVersDate,
    UserPath,AppPath,
    IniName,UserFile     : string;
    LastWidth            : integer;
    Total,Progress       : int64;
    Md5,Sha1,Sha256,
    Force,WriteToFile,
    Stopped              : boolean;
    Threads              : array  of TThreads;
    procedure SetControls (AEnable : boolean);
    procedure ShowProgress(Progresstype : TProgressType; Value : int64);
  public
    { Public-Deklarationen }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.dfm}

uses System.IniFiles, System.Hash, WinUtils, MsgDialogs, StringUtils, PathUtils,
  XlFileUtils, FileUtils, NumberUtils, GnuGetText, LangUtils, ShellDirDlg,
  InitProg, FileConsts, CalcSigDlg;

const
  wName = 200;
  wSize = 80;
  wMD5 = 160;
  wSHA1 = 190;
  wSHA256 = 190;
  wCols : array [0..4] of integer = (wName,wSize,wMD5,wSHA1,wSHA256);
  wSigs = wMD5+wSHA1+wSHA256;

{ ------------------------------------------------------------------- }
function GetCPUCount : cardinal;
var
  SysInfo : TSystemInfo;
begin
  GetSystemInfo(SysInfo);
  Result:=SysInfo.dwNumberOfProcessors;
  end;

{ ------------------------------------------------------------------- }
constructor TFileSigs.Create (AFilename : string; ASize : int64);
var
  st : TSigType;
begin
  inherited Create;
  Filename:=AFilename; Size:=ASize;
  for st:=Low(TSigType) to High(TSigType) do Signatures[st]:=''
  end;

procedure TFileSigs.SetSignature( AType : TSigType; ASig : string);
begin
  Signatures[AType]:=ASig;
  end;

{ ------------------------------------------------------------------- }
constructor TDigiSigThread.Create (const AFilename : string; ASignatureType : TSigType; ASuspend : Boolean);
begin
  inherited Create (ASuspend);
  FreeOnTerminate:=false;
  FUserBreak:=false;
  FFilename:=AFilename;
  FActiveSig:=ASignatureType;
  FSignature:='';
  FSize:=0; FPosition:=0;
  FOnProgress:=nil;
  end;

procedure TDigiSigThread.CancelThread;
begin
  FUserBreak:=true;
  end;

function TDigiSigThread.GetPosition : integer;
begin
  if FSize=0 then Result:=0
  else Result:=round(100*FPosition/FSize);
  end;

{procedure TDigiSigThread.SetProgress (ACallBack : TProgressEvent);
begin
  FOnProgress:=ACallBack;
  end;     }

function TDigiSigThread.GetDone : boolean;
begin
  Result:=Terminated;
  end;

procedure TDigiSigThread.UpdateProgress;
begin
  FOnProgress(FProgresstype,FProgressValue);
  end;

procedure TDigiSigThread.DoProgress(Progresstype : TProgressType; Value : int64);
begin
  if ProgressType=ptStart then begin
    FSize:=Value; FPosition:=0;
    end
  else if ProgressType=ptPos then FPosition:=Value;
  if assigned(FOnProgress) then begin
    FProgresstype:=Progresstype; FProgressValue:=Value;
//    FOnProgress(FProgresstype,FProgressVaue);
    Synchronize(UpdateProgress);
    end;
  end;

function TDigiSigThread.OpenFile : integer;
begin
  if FileExistsXL(FFilename) then begin
    try
      FStream:=TFileStream.Create(FilenameToXL(FFilename),fmOpenRead+fmShareDenyNone);
      Result:=errOK;
    except
      on EFOpenError do begin
        Result:=errFileOpen; Exit;
        end;
      end;
    end
  else Result:=errNotFound;
  end;

function TDigiSigThread.ProcessFile : integer;
begin
  Result:=errOK;
  end;

procedure TDigiSigThread.Execute;
begin
  FError:=ProcessFile;
  Terminate;
  end;

constructor TMD5Thread.Create (const AFilename : string; ASuspend : Boolean);
begin
  inherited Create(AFileName,stMD5,ASuspend);
  end;

function TMD5Thread.ProcessFile : integer;
var
  HashMD5  : THashMD5;
  Buffer   : array of byte;
  BufSize,BSize : integer;
begin
  Result:=OpenFile;
  if Result=errOK then begin
    HashMD5:=THashMD5.Create;
    BufSize:=defBufSize;
    SetLength(Buffer,BufSize);
    try
      with FStream do begin
        DoProgress(ptStart,Size);
        while (Position<Size) and not FUserBreak do begin
          BSize:=Read(Buffer[0],BufSize);
          DoProgress(ptInc,BSize); DoProgress(ptPos,Position);
          if BSize>0 then HashMD5.Update(Buffer[0],BSize);
          end;
        DoProgress(ptEnd,Size);
        end;
      if FUserBreak then Result:=errUserBreak
      else Result:=errOK;
    finally
      FStream.Free;
      Buffer:=nil;
      end;
    FSignature:=HashMD5.HashAsString;
    end;
  end;

procedure TMD5Thread.CancelThread;
begin
  FUserBreak:=true;
  end;

constructor TSHA1Thread.Create (const AFilename : string; ASuspend : Boolean);
begin
  inherited Create(AFileName,stSHA1,ASuspend);
  end;

function TSHA1Thread.ProcessFile : integer;
var
  HashSHA1  : THashSHA1;
  Buffer    : array of byte;
  BufSize,BSize : integer;
begin
  Result:=OpenFile;
  if Result=errOK then begin
    HashSHA1:=THashSHA1.Create;
    BufSize:=defBufSize;
    SetLength(Buffer,BufSize);
    try
      with FStream do begin
        DoProgress(ptStart,Size);
        while (Position<Size) and not FUserBreak do begin
          BSize:=Read(Buffer[0],BufSize);
          DoProgress(ptInc,BSize); DoProgress(ptPos,Position);
          if BSize>0 then HashSHA1.Update(Buffer[0],BSize);
          end;
        DoProgress(ptEnd,Size);
        end;
      if FUserBreak then Result:=errUserBreak
      else Result:=errOK;
    finally
      FStream.Free;
      Buffer:=nil;
      end;
    FSignature:=HashSHA1.HashAsString;
    end;
  end;

procedure TSHA1Thread.CancelThread;
begin
  FUserBreak:=true;
  end;

constructor TSHA256Thread.Create (const AFilename : string; ASuspend : Boolean);
begin
  inherited Create(AFileName,stMD5,ASuspend);
  end;

function TSHA256Thread.ProcessFile : integer;
var
  HashSHA2 : THashSHA2;
  Buffer   : array of byte;
  BufSize,BSize : integer;
begin
  Result:=OpenFile;
  if Result=errOK then begin
    HashSHA2:=THashSHA2.Create(SHA256);
    BufSize:=defBufSize;
    SetLength(Buffer,BufSize);
    try
      with FStream do begin
        DoProgress(ptStart,Size);
        while (Position<Size) and not FUserBreak do begin
          BSize:=Read(Buffer[0],BufSize);
          DoProgress(ptInc,BSize); DoProgress(ptPos,Position);
          if BSize>0 then HashSHA2.Update(Buffer[0],BSize);
          end;
        DoProgress(ptEnd,Size);
        end;
      if FUserBreak then Result:=errUserBreak
      else Result:=errOK;
    finally
      FStream.Free;
      Buffer:=nil;
      end;
    FSignature:=HashSHA2.HashAsString;
    end;
  end;

procedure TSHA256Thread.CancelThread;
begin
  FUserBreak:=true;
  end;

{ ------------------------------------------------------------------- }
const
  IniExt = 'ini';

  CfgSekt  = 'Config';
  SourceSekt = 'Source';
  DestSekt = 'Dest';
  FileSekt = 'Files';

  iniTop    = 'Top';
  iniLeft   = 'Left';
  iniHeight = 'Height';
  iniWidth  = 'Width';
  iniSubDir = 'Subdirectories';
  iniSepFil = 'SeparateFiles';
  iniOutput = 'OutFile';
  iniMd5    = 'MakeMd5';
  iniSha1   = 'MakeSHA1';
  iniSha2   = 'MakeSHA256';
  iniPrio   = 'Priority';

procedure TMainForm.FormCreate(Sender: TObject);
var
  i : integer;
  st : TSigType;
  s  : string;
begin
  TranslateComponent(self);
  InitPaths(AppPath,UserPath);
  InitVersion(Caption,Vers,CopRgt,3,3,ProgVersName,ProgVersDate);
  IniName:=Erweiter(AppPath,PrgName,IniExt);
  with TIniFile.Create(IniName) do begin
    Left:=ReadInteger (CfgSekt,IniLeft,50);
    Top:=ReadInteger (CfgSekt,IniTop,50);
    ClientHeight:=ReadInteger(CfgSekt,iniHeight,ClientHeight);
    ClientWidth:=ReadInteger(CfgSekt,iniWidth,ClientWidth);
    cbSubDir.Checked:=ReadBool(CfgSekt,iniSubDir,false);
    cbSepFile.Checked:=ReadBool(CfgSekt,iniSepFil,false);
    cbOut.Checked:=ReadBool(CfgSekt,iniOutput,true);
    cbMd5.Checked:=ReadBool(CfgSekt,iniMd5,true);
    cbSha1.Checked:=ReadBool(CfgSekt,iniSha1,true);
    cbSha256.Checked:=ReadBool(CfgSekt,iniSha2,true);
    cbPriority.ItemIndex:=ReadInteger(CfgSekt,iniPrio,0);
    Free;
    end;
  SourceDirEdit.LoadFromIni (IniName,SourceSekt);
  SourceDirEdit.ItemIndex:=0;
  FileEdit.LoadFromIni (IniName,FileSekt);
  FileEdit.ItemIndex:=0;
  OutFileEdit.LoadFromIni (IniName,DestSekt);
  OutFileEdit.ItemIndex:=0;
  Stopped:=true;
  with btnProcess do begin
    ImageList.GetBitmap(0,Glyph);
    Caption:=_('Start processing files');
    end;
  with udThreads do begin
    Max:=GetCPUCount;
    Position:=Max;
    SetLength(Threads,Max);
    end;
  for i:=0 to High(Threads) do with Threads[i] do begin
    FilePath:='';
    FileIndex:=-1; StatusIndex:=-1;
    Active:=false;
    for st:=Low(TSigType) to High(TSigType) do SigThread[st]:=nil;
    end;
  LastWidth:=0;
  Force:=false; WriteToFile:=false; Md5:=false; Sha1:=false; Sha256:=false;
  if ParamCount>0 then for i:=1 to ParamCount do begin
    s:=ParamStr(i);
    if s[1]='/' then begin
      Delete(s,1,1);
      if CompareOption(s,'force') then Force:=true;
      if CompareOption(s,'out') then WriteToFile:=true;
      if CompareOption(s,'md5') then Md5:=true;
      if CompareOption(s,'sha1') then Sha1:=true;
      if CompareOption(s,'sha256') then Sha256:=true;
      end
    else UserFile:=s;
    end
  else UserFile:='';
  end;

procedure TMainForm.FormResize(Sender: TObject);
var
  i,n,w : integer;
begin
  with lvFiles do begin
    n:=Width-wName-wSize-25;
    w:=0;
    for i:=2 to Columns.Count-1 do w:=w+Columns[i].Tag;
    for i:=2 to Columns.Count-1 do with Columns[i] do Width:=round(Tag*n/w);
    end;
  SourceDirEdit.SelLength:=0;
  FileEdit.SelLength:=0;
  OutFileEdit.SelLength:=0;
  end;

procedure TMainForm.FormShow(Sender: TObject);
var
  n : int64;
begin
  if length(UserFile)>0 then begin
    with CalcSigDialog do begin
      laVers.Caption:=GetVersion(3);
      if Md5 or Sha1 or Sha256 then begin
        cbMd5D.Checked:=Md5;
        cbShaD.Checked:=Sha1;
        cbSha2D.Checked:=Sha256;
        end
      else begin
        cbMd5D.Checked:=cbMd5.Checked;
        cbShaD.Checked:=cbSha1.Checked;
        cbSha2D.Checked:=cbSha256.Checked;
        end;
      with cbWriteFile do if Force then Checked:=WriteToFile
      else Checked:=WriteToFile or cbOut.Checked;
      with btnProcess do begin
        Glyph:=nil;
        ImageList.GetBitmap(0,Glyph);
        Caption:=_('Start processing files');
        end;
      edFile.Text:=UserFile;
      edSigFile.Text:=NewExt(ExtractFilename(UserFile),TxtExt);
      n:=LongFileSize(edFile.Text);
      edSize.Text:=SizeToStr(n)+' ('+IntToDecimal(n,0,3)+
         ' = $'+IntToHexStr(n,0,2)+')';
      edMd5.Text:='';
      edSHA.Text:='';
      AutoCalc:=Force;
      ShowModal;
      cbMd5.Checked:=cbMd5D.Checked;
      cbSha1.Checked:=cbShaD.Checked;
      cbOut.Checked:=cbWriteFile.Checked;
      end;
    Close;
    end;
  end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  with TIniFile.Create(IniName) do begin
    try
      WriteInteger (CfgSekt,IniLeft,Left);
      WriteInteger (CfgSekt,IniTop,Top);
      WriteInteger(CfgSekt,iniHeight,ClientHeight);
      WriteInteger(CfgSekt,iniWidth,ClientWidth);
      WriteBool(CfgSekt,iniOutput,cbOut.Checked);
      WriteBool(CfgSekt,iniSubDir,cbSubDir.Checked);
      WriteBool(CfgSekt,iniSepFil,cbSepFile.Checked);
      WriteBool(CfgSekt,iniMd5,cbMd5.Checked);
      WriteBool(CfgSekt,iniSha1,cbSha1.Checked);
      WriteBool(CfgSekt,iniSha2,cbSha256.Checked);
      WriteInteger(CfgSekt,iniPrio,cbPriority.ItemIndex);
    finally
      Free;
      end;
    end;
  end;

{ ------------------------------------------------------------------- }
procedure TMainForm.EndeBtnClick(Sender: TObject);
begin
  Close;
  end;

procedure TMainForm.InfoBtnClick(Sender: TObject);
begin
  InfoDialog(ProgName,ProgVersName+' - '+ProgVersDate+sLineBreak+
           VersInfo.CopyRight+sLineBreak+'E-Mail: '+EmailAdr);
end;

procedure TMainForm.SetControls (AEnable : boolean);
var
  i : integer;
begin
  with pnSettings do for i:=0 to ControlCount-1 do Controls[i].Enabled:=AEnable;
  edThreads.Enabled:=AEnable;
  udThreads.Enabled:=AEnable;
  end;

{ ------------------------------------------------------------------- }
procedure TMainForm.SourceSelectBtnClick(Sender: TObject);
var
  s : string;
begin
  s:=SourceDirEdit.Text;
  if length(s)=0 then s:=UserPath;
  if ShellDirDialog.Execute (_('Directory with files to process'),true,true,false,UserPath,s) then begin
    with SourceDirEdit do begin
      Text:=s; AddItem(s);
      end;
    with FileEdit do begin
      Text:='*.*'; AddItem(Text);
      end;
    end;
  end;

procedure TMainForm.FileBtnClick(Sender: TObject);
var
  i : integer;
  s : string;
begin
  with OpenDialog do begin
    InitialDir:=SourceDirEdit.Text;
    Filename:='';
    Filter:=_('all files|*.*');
    Title:=_('Select files');
    if Execute then begin
      cbSubDir.Checked:=false;
      s:=ExtractFilePath(Files.Strings[0]);
      with SourceDirEdit do if s<>Text then begin
        Text:=s; AddItem(s);
        end;
      with Files do begin   // von Opendialog
        s:='';
        for i:=0 to Count-1 do s:=s+ExtractFilename(Strings[i])+';';
        system.delete (s,length(s),1);  // remove ;
        if Count=1 then with OutFileEdit do begin
          Text:=NewExt(ExtractFilename(s),TxtExt);
          AddItem(Text);
          end;
        end;
      with FileEdit do begin
        Text:=s; AddItem(s);
        end;
      end;
    end;
  end;

procedure TMainForm.FileEditCloseUp(Sender: TObject);
var
  s,t : string;
  n : integer;
begin
  with FileEdit do s:=Items[ItemIndex]; t:=s; n:=0;
  if (length(s)=0) or (Pos('*',s)>0) or (Pos('?',s)>0) then Exit;
  repeat
    ReadNxtStr(s,';'); inc(n);
    until length(s)=0;
  if n=1 then with OutFileEdit do begin
    Text:=NewExt(t,TxtExt);
    AddItem(Text);
    end
  else OutBtnClick(Sender);
  end;

procedure TMainForm.OutBtnClick(Sender: TObject);
begin
  with OutFileEdit do begin
    Text:=defOutfile;
    AddItem(Text);
    end;
  end;

procedure TMainForm.ShowProgress(ProgressType : TProgressType; Value : int64);
begin
  if ProgressType=ptInc then begin
    inc(Progress,Value);
    ProgressBar.Position:=round(Progress/Total*100);
    end
  end;

procedure TMainForm.btnProcessClick(Sender: TObject);
var
  FListe        : TStringList;
  s,t,oName     : string;
  oText         : TextFile;
  sErr,sCnt,
  i,j,k,ec,nt   : integer;
  fs            : TFileStream;
  sz            : int64;
  st            : TSigType;
  stSet         : TSigTypes;
  SigIndex      : array [TSigType] of integer;
  ok,AllDone    : boolean;
  tm            : TDateTime;
  ThPr          : TThreadPriority;
  AllSignatures : TSignatures;

  procedure ScanDir (Dir,Mask : string; var ATotal : int64);
  var
    DirInfo    : TSearchRec;
    Findresult : integer;
  begin
    if cbSubDir.Checked then begin
      FindResult:=FindFirst (Erweiter(Dir,'*','*'),faAnyFile,DirInfo);
      while FindResult=0 do with DirInfo do begin
        if NotSpecialDir(Name) then begin
          if((Attr and faDirectory)<>0) then ScanDir (Erweiter(Dir,Name,''),Mask,ATotal);
          end;
        FindResult:=FindNext (DirInfo);
        end;
      FindClose(DirInfo);
      end;
    FindResult:=FindFirst (Erweiter(Dir,Mask,''),faArchive+faReadOnly+faHidden+faSysfile,DirInfo);
    while FindResult=0 do with DirInfo do begin
      if NotSpecialDir(Name) and not AnsiSameText(oName,Name) then begin
        FListe.Add(Erweiter(Dir,Name,''));
        ATotal:=ATotal+Size;
        end;
      FindResult:=FindNext (DirInfo);
      end;
    FindClose(DirInfo);
    end;

begin
  if not Stopped then begin
    Stopped:=true; Exit;
    end;
  Stopped:=false;
  if not cbMD5.Checked and not cbSHA1.Checked and not cbSHA256.Checked then begin
    cbMD5.Checked:=true;
    end;
  with btnProcess do begin
    Glyph:=nil;
    ImageList.GetBitmap(1,Glyph);
    Caption:=_('Stop processing files');
    end;
  EndeBtn.Enabled:=false;
  SetControls(false);
  StatusBar.SimpleText:='';
  t:=FileEdit.Text;
  case cbPriority.Itemindex of
    1 : ThPr:=tpHigher;
    2 : ThPr:=tpHighest;
    3 : ThPr:=tpLower;
    else ThPr:=tpNormal;
    end;
  if length(t)>0 then begin
    lvFiles.Clear;
    FListe:=TStringList.Create;
    sErr:=0; sCnt:=0; Total:=0; Progress:=0; nt:=0; stSet:=[];
    if cbOut.Checked then oName:=OutFileEdit.Text else OName:='';
    if cbMD5.Checked or cbSHA1.Checked or cbSHA256.Checked then begin
      repeat
        s:=ReadNxtStr (t,';');
        ScanDir (SourceDirEdit.Text,s,Total);
        until length(t)=0;
      if cbMD5.Checked then inc(nt);
      if cbSHA1.Checked then inc(nt);
      if cbSHA256.Checked then inc(nt);
      Total:=nt*Total;
      end;
    with lvFiles.Columns do begin
      Clear;
      for st:=Low(TSigType) to High(TSigType) do SigIndex[st]:=-1;
      with Add do begin
        Caption:=_('Name'); Width:=wName;
        end;
      with Add do begin
        Caption:=_('Size'); Width:=wSize;
        end;
      if cbMD5.Checked then with Add do begin
        Caption:=SigNames[stMD5]; Tag:=wMD5;
        include(stSet,stMD5);
        SigIndex[stMd5]:=Index-1;
        end;
      if cbSHA1.Checked then with Add do begin
        Caption:=SigNames[stSHA1]; Tag:=wSHA1;
        include(stSet,stSHA1);
        SigIndex[stSHA1]:=Index-1;
        end;
      if cbSHA256.Checked then with Add do begin
        Caption:=SigNames[stSHA256]; Tag:=wSHA256;
        include(stSet,stSHA256);
        SigIndex[stSHA256]:=Index-1;
        end;
      end;
    nt:=udThreads.Position div nt -1;
    FormResize(Sender);
    with FListe do for i:=0 to Count-1 do
      Objects[i]:=TFileSigs.Create(MakeRelativePath(SetDirName(SourceDirEdit.Text),Strings[i]),LongFileSize(Strings[i]));
    if FListe.Count>0 then begin
      i:=0; tm:=Now;
      repeat
        ok:=true;
        for j:=0 to nt do with Threads[j] do begin
          if Active then begin
            if Stopped then begin
              for st:=Low(TSigType) to High(TSigType) do
                  if assigned(SigThread[st]) then SigThread[st].CancelThread;
              repeat
                Sleep(1);
                Application.ProcessMessages;
                AllDone:=true;
                for st:=Low(TSigType) to High(TSigType) do
                  if assigned(SigThread[st]) then AllDone:=AllDone and SigThread[st].Done;
                until AllDone;
              end;
            end;
          AllDone:=true;
          for st:=Low(TSigType) to High(TSigType) do if assigned(SigThread[st]) then with SigThread[st] do begin
            k:=SigIndex[st];
            if k>=0 then begin
              if Done then begin
                if ErrorCode=errOK then s:=UpperCase(Signature)
                else begin
                  inc(sErr); s:=GetCopyErrMsg(ErrorCode);; //_('#Error#');
                  end;
                (FListe.Objects[FileIndex] as TFileSigs).SetSignature(st,s);
                lvFiles.Items[Statusindex].SubItems.Strings[k]:=s;
                FreeAndNil(SigTHread[st]);
                end
              else begin
                AllDone:=false;
                with lvFiles.Items[StatusIndex].SubItems do begin
                  if ProcessStatus>integer(Objects[k]) then begin
                    Objects[k]:=pointer(ProcessStatus);
                    Strings[k]:=IntToStr(ProcessStatus)+' %';
                    end;
                  end;
                end;
              end;
            Active:=not AllDone;
            end;
          if not Stopped and not Active and (i<FListe.Count) then begin
            FileIndex:=i;
            sz:=LongFileSize(FListe.Strings[i]);
            inc(sCnt);
            for st:=Low(TSigType) to High(TSigType) do if st in stSet then begin
              case st of
              stMd5 : SigThread[st]:=TMd5Thread.Create(FListe.Strings[i],true);
              stSHA1 : SigThread[st]:=TSHA1Thread.Create(FListe.Strings[i],true);
              stSHA256 : SigThread[st]:=TSHA256Thread.Create(FListe.Strings[i],true);
              else SigThread[st]:=nil;
                end;
              end
            else SigThread[st]:=nil;
            inc(i);
            if (i<FListe.Count) and AnsiSameText(FListe.Strings[i],oName) then inc(i);
            Active:=true;
            for st:=Low(TSigType) to High(TSigType) do if assigned(SigThread[st]) then with SigThread[st] do begin              Priority:=ThPr;
              OnProgress:=ShowProgress;
              Start;
              FilePath:=MakeRelativePath(SetDirName(SourceDirEdit.Text),Filename);
              end;
            with lvFiles do begin
              with Items.Add do begin
                Caption:=FilePath;
                StatusIndex:=Index;
                SubItems.AddObject(IntToStr(sz),pointer(0));
                if stMD5 in stSet then SubItems.AddObject('0 %',pointer(0));
                if stSHA1 in stSet then SubItems.AddObject('0 %',pointer(0));
                if stSHA256 in stSet then SubItems.AddObject('0 %',pointer(0));
                MakeVisible(false);
                end;
              ItemIndex:=StatusIndex;
              end;
            end;
          ok:=ok and not Active;
          end;
        Sleep(100);
        Application.ProcessMessages;
        until ok;
      if length(OName)>0 then begin
        t:=SetDirName(SourceDirEdit.Text);
        AssignFile(oText,t+oName);
        rewrite(oText);
        if cbSepFile.Checked then begin
          s:='';
          with FListe do for i:=0 to Count-1 do if Assigned(Objects[i]) then
              with (Objects[i] as TFileSigs) do begin
            if not AnsiSametext(s,ExtractFilePath(Filename)) then begin
              CloseFile(oText);
              s:=ExtractFilePath(Filename);
              AssignFile(oText,t+s+oName);
              rewrite(oText);
              end;
            writeln (oText,ExtractFilename(Filename)+' ('+IntToStr(Size)+')');
            for st:=Low(TSigType) to High(TSigType) do if length(Signatures[st])>0 then
              writeln (oText,AddSp(SigNames[st],7)+': ',Signatures[st]);
            writeln (oText);
            end;
          end
        else begin
          with FListe do for i:=0 to Count-1 do if not AnsiSameText(Strings[i],oName) and Assigned(Objects[i]) then
              with (Objects[i] as TFileSigs) do begin
            writeln (oText,Filename+' ('+IntToStr(Size)+')');
            for st:=Low(TSigType) to High(TSigType) do if length(Signatures[st])>0 then
              writeln (oText,AddSp(SigNames[st],7)+': ',Signatures[st]);
            writeln (oText);
            end;
          end;
        CloseFile(oText);
        end;
      end
    else ErrorDialog(BottomRightPos(btnProcess),_('File(s) not found!'));
    s:=' ('+TimeToStr(Now-tm)+')';
    ProgressBar.Position:=100;
    with StatusBar do begin
      if sCnt=0 then SimpleText:=_( 'No files processed')
      else if sCnt=1 then SimpleText:=_(' 1 file processed')+s
      else SimpleText:=Format(_(' %u Files processed'),[sCnt])+s;
      if sErr>0 then SimpleText:=SimpleText+' - '+IntTostr(sErr)+_(' Errors');
      end;
    FreeListObjects(FListe);
    FListe.Free;
    end
  else StatusBar.SimpleText:=_(' Nothing to do!');
  Application.ProcessMessages;
  Stopped:=true;
  with btnProcess do begin
    Glyph:=nil;
    ImageList.GetBitmap(0,Glyph);
    Caption:=_('Start processing files');
    end;
  EndeBtn.Enabled:=true;
  SetControls(true);
  end;

end.
